/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "verb.h"

#include <stdlib.h>
#include <string.h>

#include "verb_gui.h"
#include "sverb.h"

static float do_verb_left(verb *this, float in)
{
  float out = this->s_left.sverb_process(&this->s_left, in);
  this->l_delay[this->left_cur] = out;
  this->left_cur++;
  this->left_cur &= 255;
  return this->l_delay[(this->left_cur - this->left_delay - 1) & 255] *
         this->level_reverb + in * this->level_signal;
}

static float do_verb_right(verb *this, float in)
{
  float out = this->s_right.sverb_process(&this->s_right, in);
  this->r_delay[this->right_cur] = out;
  this->right_cur++;
  this->right_cur &= 255;
  return this->r_delay[(this->right_cur - this->right_delay - 1) & 255] *
         this->level_reverb + in * this->level_signal;
}

static int feed_buffer_float_mono(verb *this, float *buf, int nb_samples, int start)
{
  int i;
  int sub_nb_samples = nb_samples;
  int me_nb_samples = 0;

  if (start > this->last_point) return 0;

  /* the reverb effect adds samples to let the reverberation finish,
   * we must take care of not asking too much data to the child
   */
  if (start + nb_samples - 1 > this->sub_last_point) {
    sub_nb_samples = this->sub_last_point - start + 1;
    /* maybe it's < 0 */
    if (sub_nb_samples < 0) sub_nb_samples = 0;
    if (sub_nb_samples > nb_samples) sub_nb_samples = nb_samples;
    me_nb_samples = nb_samples - sub_nb_samples;
  }

  if (this->c.bypass) {
    /* set to 0 supplementary samples */
    if (me_nb_samples)
      memset(buf+sub_nb_samples, 0, me_nb_samples * sizeof(float));
    return this->c.subchannel->c.feed_buffer_float_mono(this->c.subchannel,
        buf, sub_nb_samples, start);
  }

  if (sub_nb_samples)
    if (!this->c.subchannel->c.feed_buffer_float_mono(this->c.subchannel,
          buf, sub_nb_samples, start))
      return 0;

  if (!this->s_left.delays[0] && !this->s_left.delays_int[0]) return 1;

  for (i = 0; i < sub_nb_samples; i++)
    buf[i] = do_verb_left(this, buf[i]);

  for (i = 0; i < me_nb_samples; i++)
    buf[i+sub_nb_samples] = do_verb_left(this, 0);

  return 1;
}

static int feed_buffer_float_stereo(verb *this, float *buf, int nb_samples, int start)
{
  int i;
  int sub_nb_samples = nb_samples;
  int me_nb_samples = 0;

  if (start > this->last_point) return 0;

  /* the reverb effect adds samples to let the reverberation finish,
   * we must take care of not asking too much data to the child
   */
  if (start + nb_samples - 1 > this->sub_last_point) {
    sub_nb_samples = this->sub_last_point - start + 1;
    /* maybe it's < 0 */
    if (sub_nb_samples < 0) sub_nb_samples = 0;
    if (sub_nb_samples > nb_samples) sub_nb_samples = nb_samples;
    me_nb_samples = nb_samples - sub_nb_samples;
  }

  if (this->c.bypass) {
    /* set to 0 supplementary samples */
    if (me_nb_samples)
      memset(buf+sub_nb_samples*2, 0, me_nb_samples * 2 * sizeof(float));
    return this->c.subchannel->c.feed_buffer_float_stereo(this->c.subchannel,
        buf, sub_nb_samples, start);
  }

  if (sub_nb_samples)
    if (!this->c.subchannel->c.feed_buffer_float_stereo(this->c.subchannel,
          buf, sub_nb_samples, start))
      return 0;

  if (!this->s_left.delays[0] && !this->s_left.delays_int[0]) return 1;

  for (i = 0; i < sub_nb_samples * 2; i++) {
    buf[i] = do_verb_left(this, buf[i]);
    i++;
    buf[i] = do_verb_right(this, buf[i]);
  }

  for (i = 0; i < me_nb_samples * 2; i++) {
    buf[i+sub_nb_samples*2] = do_verb_left(this, 0);
    i++;
    buf[i+sub_nb_samples*2] = do_verb_right(this, 0);
  }

  return 1;
}

static int first_point(verb *this)
{
  return this->c.subchannel->c.first_point(this->c.subchannel);
}

static int last_point(verb *this)
{
  int i;
  int last_point = this->c.subchannel->c.last_point(this->c.subchannel);

  /* remember kid's last_point */
  this->sub_last_point = last_point;

  /* we've got to add the reverb delay lengths
   * (well, not sure if it is correct or not...)
   */
  for (i = 0; i < 15; i++)
    last_point += (float)this->s_left.DELAY[i] / 0.15;

  /* remember our own last_point */
  this->last_point = last_point;

  return last_point;
}

static void reset(verb *this)
{
  int i;

  this->c.subchannel->c.reset(this->c.subchannel);

  save_reverb_from_gui(this);
  if (this->c.gui_on)
    from_verb_to_gui(this);

  sverb_reset(&this->s_left);
  sverb_reset(&this->s_right);

  for (i = 0; i < 256; i++) {
    this->l_delay[i] = 0;
    this->r_delay[i] = 0;
  }
  this->left_cur = 0;
  this->right_cur = 0;

  /* we need the last_point of the child (and ours as well) */
  last_point(this);
}

verb *new_verb(channel *c)
{
  verb *ret = malloc(sizeof(verb));

  if (!ret)
    return 0;

  memset(ret, 0, sizeof(verb));

  ret->c.type = CHANNEL_TYPE_VERB;

  ret->c.subchannel = c;

  ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))feed_buffer_float_mono;
  ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))feed_buffer_float_stereo;
  ret->c.first_point = (int (*)(void *))first_point;
  ret->c.last_point = (int (*)(void *))last_point;
  ret->c.reset = (void (*)(void *))reset;

  ret->c.open_gui = (void (*)(void *, int))verb_open_gui;
  ret->c.close_gui = (void (*)(void *))verb_close_gui;

  sverb_init(&ret->s_left);
  sverb_init(&ret->s_right);
  set_verb(ret, SVERB_AVERAGE_1);

  return ret;
}

void delete_verb(verb *e)
{
  sverb_free(&e->s_left);
  sverb_free(&e->s_right);
  free(e);
}
